from jinja2 import Template

from markupy import li


def test_template_injection() -> None:
    result = Template("<ul>{{ stuff }}</ul>").render(stuff=li["I am safe!"])
    assert result == "<ul><li>I am safe!</li></ul>"



import pytest
from markupsafe import Markup

import markupy
from markupy import MarkupyElement, div


def test_instance_cache() -> None:
    """
    markupy creates element object dynamically. make sure they are reused and case insensitive
    """
    assert markupy.div is markupy.div
    assert markupy.div is markupy.Div
    assert markupy.div is not markupy.div()
    assert markupy.div is not markupy.div[None]


def test_element_repr() -> None:
    assert repr(markupy.div("#a")) == """<MarkupyElement '<div id="a"></div>'>"""


def test_void_element_repr() -> None:
    assert repr(markupy.hr("#a")) == """<MarkupyElement '<hr id="a">'>"""


def test_markup_str() -> None:
    result = str(div(id="a"))
    assert isinstance(result, str)
    assert isinstance(result, Markup)
    assert result == '<div id="a"></div>'


def test_element_type() -> None:
    assert type(div) is MarkupyElement
    assert type(div(a="test")) is MarkupyElement
    assert type(div["a"]) is MarkupyElement
    assert type(div()["a"]) is MarkupyElement



import re
from collections.abc import Iterable
from typing import TypeAlias

from markupsafe import escape

AttributeValue: TypeAlias = None | bool | str | int


def _classes_to_str(classes: Iterable[str]) -> str:
    return " ".join(filter(None, classes))


def _rewrite_attr_key(key: str) -> str:
    # First char "_" not followed by another "_" -> "@"
    key = re.sub(r"^(?:_)([^_]+)", r"@\1", key)
    # Trailing underscore "_" is meaningless and is used to escape protected
    # keywords that might be used as attr keys such as class_ and for_
    key = key.removesuffix("_")
    # Upper case -> "-"
    key = "-".join(filter(None, re.sub(r"([A-Z])", r" \1", key).split())).lower()
    # Double underscore -> ":"
    key = key.replace("__", ":")
    # Single underscore -> "."
    key = key.replace("_", ".")
    return key


class AttributeDict(dict):
    def __setitem__(self, key: str, value: AttributeValue) -> None:
        if not isinstance(key, str):
            raise ValueError("Attribute key must be a string")
        if not isinstance(value, AttributeValue):
            raise ValueError(f"Invalid `{key}` attribute type `{value!r}`")
        if value is False or value is None:
            return

        return super().__setitem__(key, value)

    def __str__(self) -> str:
        pairs = filter(lambda pair: pair[0] and pair[1], self.items())
        return " ".join(
            escape(str(k)) if v is True else f'{escape(str(k))}="{escape(str(v))}"'
            for k, v in pairs
        )

    def add_selector(self, selector: str) -> None:
        if selector is None:
            return
        if not isinstance(selector, str):
            raise ValueError(f"Selector string must be a str, got {selector!r}")

        match = re.match(r"^(#.+?)?(\..+?)?$", selector)
        if match is None:
            raise ValueError("Selector string format is invalid")

        if id_str := match.group(1):
            self["id"] = id_str[1:]
        if cls_str := match.group(2):
            classes = cls_str[1:].split(".")
            self["class"] = _classes_to_str(classes)

    def add_dict(self, dct: dict, *, rewrite_keys: bool = False) -> None:
        if dct is None:
            return
        if not isinstance(dct, dict):
            raise ValueError(f"Attributes must be provided as a dict, got {dct!r}")
        for key, value in dct.items():
            if key != "_" and rewrite_keys:
                # Preserve single _ for hyperscript
                key = _rewrite_attr_key(key)

            if key == "class" and not isinstance(value, str):
                if isinstance(value, dict):
                    # Check for dict before more generic Iterable
                    # Drop all False-ish class values
                    classes = [k for k in value if value[k]]
                    value = _classes_to_str(classes)
                elif isinstance(value, Iterable):
                    value = _classes_to_str(value)

            self[key] = value



import re as _re

from markupsafe import Markup as _Markup

from .element import MarkupyElement, Node, iter_node


def render_node(node: Node) -> _Markup:
    return _Markup("".join(iter_node(node)))


def __getattr__(name: str) -> MarkupyElement:
    # Consider uppercase chars and underscores as word boundaries for tag names
    words = filter(None, _re.sub(r"([A-Z])", r"_\1", name).split("_"))
    html_name = "-".join(words).lower()
    return MarkupyElement.by_name(html_name)



import functools
from collections.abc import Callable, Iterable, Iterator
from typing import Any, Self, TypeAlias, overload

from markupsafe import Markup, escape

from .attribute import AttributeDict, AttributeValue


class MarkupyElement:
    def __init__(self, name: str) -> None:
        self.name = name
        self.attributes = AttributeDict()
        self.children = None

    def is_void(self) -> bool:
        return self.name in {
            "area",
            "base",
            "br",
            "col",
            "command",
            "embed",
            "hr",
            "img",
            "input",
            "keygen",
            "link",
            "meta",
            "param",
            "source",
            "track",
            "wbr",
        }

    def __iter__(self) -> Iterator[str]:
        if self.name == "html":
            yield "<!doctype html>"

        opening = f"{self.name} {self.attributes}".strip()
        yield f"<{opening}>"
        if not self.is_void():
            yield from iter_node(self.children)
            yield f"</{self.name}>"

    def __str__(self) -> str:
        return Markup("".join(str(x) for x in self))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self}'>"

    @classmethod
    @functools.lru_cache(maxsize=300)
    def by_name(cls, name: str) -> Self:
        return cls(name)

    def to_element(self) -> "MarkupyElement":
        # When imported, elements are loaded from cache
        # Make sure we re-instantiate them on setting attributes/children
        # to avoid sharing attributes/children between multiple instances
        if self is MarkupyElement.by_name(self.name):
            return MarkupyElement(self.name)
        return self

    # Use call syntax () to define attributes
    @overload
    def __call__(
        self, selector: str, attributes: dict[str, AttributeValue], **kwargs: Any
    ) -> "MarkupyElement": ...
    @overload
    def __call__(self, selector: str, **kwargs: Any) -> "MarkupyElement": ...
    @overload
    def __call__(
        self, attributes: dict[str, AttributeValue], **kwargs: Any
    ) -> "MarkupyElement": ...
    @overload
    def __call__(self, **kwargs: Any) -> Self: ...
    def __call__(self, *args: Any, **kwargs: Any) -> "MarkupyElement":
        selector = None
        attributes_dict = None
        attributes_kwargs = kwargs
        if len(args) == 1:
            if isinstance(args[0], str):
                # element(".foo")
                selector = args[0]
            else:
                # element({"foo": "bar"})
                attributes_dict = args[0]
        elif len(args) == 2:
            # element(".foo", {"bar": "baz"})
            selector, attributes_dict = args

        el = self.to_element()
        try:
            el.attributes.add_selector(selector)
        except ValueError:
            raise ValueError(f"Invalid selector string `{selector}` for element {self}")
        try:
            el.attributes.add_dict(attributes_dict)
        except ValueError:
            raise ValueError(
                f"Invalid dict attributes `{attributes_dict}` for element {self}"
            )
        try:
            el.attributes.add_dict(attributes_kwargs, rewrite_keys=True)
        except ValueError:
            raise ValueError(
                f"Invalid keyword attributes `{attributes_kwargs}` for element {self}"
            )
        return el

    # Use subscriptable [] syntax to assign children
    def __getitem__(self, children: "Node") -> "MarkupyElement":
        if self.is_void():
            raise ValueError(f"Void element {self} cannot contain children")

        el = self.to_element()
        el.children = children
        return el

    # Allow starlette Response.render to directly render this element without
    # explicitly casting to str:
    # https://github.com/encode/starlette/blob/5ed55c441126687106109a3f5e051176f88cd3e6/starlette/responses.py#L44-L49
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes:
        return str(self).encode(encoding, errors)

    # Avoid having Django "call" a markupy element that is injected into a
    # template. Setting do_not_call_in_templates will prevent Django from doing
    # an extra call:
    # https://docs.djangoproject.com/en/5.0/ref/templates/api/#variables-and-lookups
    do_not_call_in_templates = True


Node: TypeAlias = (
    None | bool | str | int | MarkupyElement | Iterable["Node"] | Callable[[], "Node"]
)


def iter_node(node: Node) -> Iterator[str]:
    if node is None or isinstance(node, bool):
        return
    while not isinstance(node, MarkupyElement) and callable(node):
        node = node()

    if isinstance(node, MarkupyElement):
        yield from node
    elif isinstance(node, int):
        yield str(node)
    elif isinstance(node, str):
        yield str(escape(node))
    elif isinstance(node, Iterable) and not isinstance(node, bytes):
        for child in node:
            yield from iter_node(child)
    else:
        raise ValueError(f"{node!r} is not a valid child element")



import typing as t

import pytest
from markupsafe import Markup

from markupy import Button, Input, button, div, th


def test_attribute() -> None:
    assert str(div(id="hello")["hi"]) == '<div id="hello">hi</div>'


class Test_class_names:
    def test_str(self) -> None:
        result = div(class_='">foo bar')
        assert str(result) == '<div class="&#34;&gt;foo bar"></div>'

    def test_safestring(self) -> None:
        result = div(class_=Markup('">foo bar'))
        assert str(result) == '<div class="&#34;&gt;foo bar"></div>'

    def test_list(self) -> None:
        result = div(class_=['">foo', Markup('">bar'), False, None, "", "baz"])
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    def test_tuple(self) -> None:
        result = div(class_=('">foo', Markup('">bar'), False, None, "", "baz"))
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    def test_dict(self) -> None:
        result = div(
            class_={'">foo': True, Markup('">bar'): True, "x": False, "baz": True}
        )
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    # def test_nested_dict(self) -> None:
    #     result = div(
    #         class_=[
    #             '">list-foo',
    #             Markup('">list-bar'),
    #             {'">dict-foo': True, Markup('">list-bar'): True, "x": False},
    #         ]
    #     )
    #     assert str(result) == (
    #         '<div class="&#34;&gt;list-foo &#34;&gt;list-bar '
    #         '&#34;&gt;dict-foo &#34;&gt;list-bar"></div>'
    #     )

    def test_false(self) -> None:
        result = str(div(class_=False))
        assert result == "<div></div>"

    def test_none(self) -> None:
        result = str(div(class_=None))
        assert result == "<div></div>"

    def test_no_classes(self) -> None:
        result = str(div(class_={"foo": False}))
        assert result == "<div></div>"


def test_dict_attributes() -> None:
    result = div({"@click": 'hi = "hello"'})

    assert str(result) == """<div @click="hi = &#34;hello&#34;"></div>"""


def test_underscore() -> None:
    # Hyperscript (https://hyperscript.org/) uses _, make sure it works good.
    result = div(_="foo")
    assert str(result) == """<div _="foo"></div>"""


def test_dict_attributes_avoid_replace() -> None:
    result = div({"class_": "foo", "hello_hi": "abc"})
    assert str(result) == """<div class_="foo" hello_hi="abc"></div>"""


def test_dict_attribute_false() -> None:
    result = div({"bool-false": False})
    assert str(result) == "<div></div>"


def test_dict_attribute_true() -> None:
    result = div({"bool-true": True})
    assert str(result) == "<div bool-true></div>"


def test_uppercase_replacement() -> None:
    result = Button(hxPost="/foo")["click me!"]
    assert str(result) == """<button hx-post="/foo">click me!</button>"""


def test_at_replacement() -> None:
    result = Input(_input_debounce_500ms="fetchResults")
    assert str(result) == """<input @input.debounce.500ms="fetchResults">"""


def test_colon_replacement() -> None:
    result = Button(
        hxOn__htmx__configRequest="event.detail.parameters.example = 'Hello Scripting!'"
    )["Post Me!"]
    assert (
        str(result)
        == """<button hx-on:htmx:config-request="event.detail.parameters.example = &#39;Hello Scripting!&#39;">Post Me!</button>"""
    )


class Test_attribute_escape:
    pytestmark = pytest.mark.parametrize(
        "x",
        [
            '<"foo',
            Markup('<"foo'),
        ],
    )

    def test_dict(self, x: str) -> None:
        result = div({x: x})
        assert str(result) == """<div &lt;&#34;foo="&lt;&#34;foo"></div>"""

    def test_kwarg(self, x: str) -> None:
        result = div(**{x: x})
        assert str(result) == """<div &lt;&#34;foo="&lt;&#34;foo"></div>"""


def test_boolean_attribute_true() -> None:
    result = button(disabled=True)
    assert str(result) == "<button disabled></button>"


def test_kwarg_attribute_none() -> None:
    result = div(foo=None)
    assert str(result) == "<div></div>"


def test_dict_attribute_none() -> None:
    result = div({"foo": None})
    assert str(result) == "<div></div>"


def test_boolean_attribute_false() -> None:
    result = button(disabled=False)
    assert str(result) == "<button></button>"


def test_integer_attribute() -> None:
    result = th(colspan=123)
    assert str(result) == '<th colspan="123"></th>'


def test_id_class() -> None:
    result = div("#myid.cls1.cls2")

    assert str(result) == """<div id="myid" class="cls1 cls2"></div>"""


def test_id_class_only_id() -> None:
    result = div("#myid")
    assert str(result) == """<div id="myid"></div>"""


def test_id_class_only_classes() -> None:
    result = div(".foo.bar")
    assert str(result) == """<div class="foo bar"></div>"""


# def test_id_class_wrong_order() -> None:
#     with pytest.raises(
#         ValueError, match="id \\(#\\) must be specified before classes \\(\\.\\)"
#     ):
#         div(".myclass#myid")


def test_id_class_bad_format() -> None:
    with pytest.raises(
        ValueError  # , match="id/class strings must start with # or ."
    ):
        div("foo")


def test_id_class_bad_type() -> None:
    with pytest.raises(
        ValueError  # , match="id/class strings must be str. got {'oops': 'yes'}"
    ):
        div({"oops": "yes"}, {})  # type: ignore


def test_id_class_and_kwargs() -> None:
    result = div("#theid", for_="hello", dataFoo="<bar")
    assert str(result) == """<div id="theid" for="hello" data-foo="&lt;bar"></div>"""


def test_attrs_and_kwargs() -> None:
    result = div({"a": "1", "for": "a"}, for_="b", b="2")
    assert str(result) == """<div a="1" for="b" b="2"></div>"""


def test_class_priority() -> None:
    result = div(".a", {"class": "b"}, class_="c")
    assert str(result) == """<div class="c"></div>"""

    result = div(".a", {"class": "b"})
    assert str(result) == """<div class="b"></div>"""


def test_attribute_priority() -> None:
    result = div({"foo": "a"}, foo="b")
    assert str(result) == """<div foo="b"></div>"""


@pytest.mark.parametrize("not_an_attr", [1234, b"foo", object(), object, 1, 0, None])
def test_invalid_attribute_key(not_an_attr: t.Any) -> None:
    with pytest.raises(ValueError):
        str(div({not_an_attr: "foo"}))


@pytest.mark.parametrize(
    "not_an_attr",
    [12.34, b"foo", object(), object],
)
def test_invalid_attribute_value(not_an_attr: t.Any) -> None:
    with pytest.raises(ValueError):
        div(foo=not_an_attr)
